r"""
===========
Simulations
===========
This module contains classes that perform simulations directly on the images
to obtain structural information

.. autoclass:: RandomWalk
    :members:

.. autoclass:: Porosimetry
    :members:

"""
from .__RandomWalk__ import RandomWalk
from .__Porosimetry__ import Porosimetry
