r'''
=============
Visualization
=============

This module contains functions for quickly visualizing 3D images in 2D views.

.. autofunction:: sem
.. autofunction:: xray
.. autofunction:: set_mpl_style


'''


from .__views__ import sem
from .__views__ import xray
from .__funcs__ import set_mpl_style
