r'''
============
Binarization
============

This module contains functions for thresholding greyscale images into binary
images (i.e. True and False) indicating the phases.

.. autofunction:: simple_otsu

'''

from .__funcs__ import simple_otsu
