r"""
=======
Metrics
=======

A suite of tools for determine key metrics about an image.  Typically these are
applied to an image after applying a filter, but a few functions can be applied
directly to the binary image.

.. autofunction:: chord_length_counts
.. autofunction:: chord_length_distribution
.. autofunction:: pore_size_density
.. autofunction:: pore_size_distribution
.. autofunction:: porosity
.. autofunction:: representative_elementary_volume
.. autofunction:: two_point_correlation_bf
.. autofunction:: two_point_correlation_fft

"""

from .__funcs__ import chord_length_counts
from .__funcs__ import chord_length_distribution
from .__funcs__ import pore_size_density
from .__funcs__ import pore_size_distribution
from .__funcs__ import porosity
from .__funcs__ import representative_elementary_volume
from .__funcs__ import two_point_correlation_bf
from .__funcs__ import two_point_correlation_fft
