r"""
==========
Generators
==========

This module contains a variety of functions for generating artificial images
of porous materials, generally for testing, validation, debugging, and
illustration purposes.

.. autofunction:: add_noise
.. autofunction:: blobs
.. autofunction:: bundle_of_tubes
.. autofunction:: circle_pack
.. autofunction:: cylinders
.. autofunction:: insert_shape
.. autofunction:: line_segment
.. autofunction:: noise
.. autofunction:: norm_to_uniform
.. autofunction:: overlapping_spheres
.. autofunction:: polydisperse_spheres
.. autofunction:: sphere_pack
.. autofunction:: voronoi_edges

"""

from .__imgen__ import add_noise
from .__imgen__ import blobs
from .__imgen__ import bundle_of_tubes
from .__imgen__ import circle_pack
from .__imgen__ import cylinders
from .__imgen__ import insert_shape
from .__imgen__ import line_segment
from .__imgen__ import noise
from .__imgen__ import norm_to_uniform
from .__imgen__ import overlapping_spheres
from .__imgen__ import polydisperse_spheres
from .__imgen__ import sphere_pack
from .__imgen__ import voronoi_edges
